#include <avr/io.h>
#include <avr/interrupt.h>


// для теста
byte ii;

#define ZERO_CROSS       2  //номер пина сигнала детектора нуля (дискретный вход)
#define TRIAC_GATE_PULSE 22   //номер пина сигнала управления триаком (дискретный выход)
#define VOLTAGE_ADC      3   //номер аналогового пина измерения мгновенного напряжения 
                             //при переназначении необходимо изменить значение ADMUX в разделе setup

#define widthTRIAC_GATE_PULSE 1   //длительность импульса управления триаком в тиках таймера3
                                  //при делителе таймера3 = 256 (16МГц/256=62500Гц = 16мкс/тик)
                                  //длительность импульса = 4*16мкс=64мкс
                                  
#define samples_frequencyVOLTAGE_MEASUREMENT 10    // период выборки значений напряжения в тиках Таймера1 для 62-х выборок
                                                   //в течение полуволны переменного напряжения
                                                   // Измерения запускаются по прерыванию переполнения Таймера 1
                                                   // Частота полуволны переменнного напряжения 50Гц составляет 100Гц 
                                                   // В течение полуволны Таймер1 выдает 62500Гц/100Гц=625 тиков (10мс)
                                                   // Для обеспечения 62 выборок в течение полуволны выборка должна запускаться
                                                   // c периодом 625/60=10 тик (160 мксек)
                                                   

byte statusVOLTAGE_MEASUREMENT = 1; //0-не активное, 1-запрос на начало выборки, 2-в процессе выборки (поиска максимального амплитудного) и установка 
                                    //измерения вызываются ХХ раз в секунду, измеряются значения в пределе полупериода и вычисляется максимальное амплитудное значение
                                    //частота выборки в пределе полуволны определятется значением компаратора счетчика OCR1B
                                    //старт измерений происходит по прерыванию перехода напряжения через 0

// Алгоримт расчета задержки включения триака
// * Измеряем максимальное амплитудное значение напряжения (Umax_adc) в течение полуволны в единицах АЦП 
unsigned int maxADC_AMPLITUDE_VOLTAGE;
// * Переводим полученное значение в вольты путем умножения на коэф. Z (определяется путем калибровки)
float max_volt_AMPLITUDE_VOLTAGE = 0; //используется только во время калбировки
float Z = 1/2.5;
float heaterRESISTANCE = 645; //сопротивление наргевателя в Омах, значение для лампочки 75 Вт
// * Определяем среднедействующее напряжение по упрощенной формуле Urms = Z*Umax/sqrt(2)
// * Определяем текущую мощность Ptek=Urms*Urms/R=Z*maxADC_AMPLITUDE_VOLTAGE*Z*maxADC_AMPLITUDE_VOLTAGE/(2*heaterRESISTANCE)
// * Рассчитываем коэффициент К соотношения заданной мощности (Pzad)к текущей K=Pzad/Ptek= 2*heaterRESISTANCE/(Z*Z*maxADC_AMPLITUDE_VOLTAGE*maxADC_AMPLITUDE_VOLTAGE)*requiredPOWER
//   K=(2*heaterRESISTANCE/(Z*Z))*requiredPOWER/(maxADC_AMPLITUDE_VOLTAGE*maxADC_AMPLITUDE_VOLTAGE)
//   для сокращение операций деления первую часть формулы расчитываем один раз как константу CONST=2*heaterRESISTANCE/(Z*Z)
//   тогда формула приобретае вид К = CONST*requiredPOWER/(maxADC_AMPLITUDE_VOLTAGE*maxADC_AMPLITUDE_VOLTAGE)
float requiredPOWER = 0; //заданная мощность
unsigned int K; //коэффициент соотношения заданной мощности к текущей
float CONST;
// * В таблице зависимости коэффициента К соотношения заданной мощности к текущей от задержки открытия триака определяем задержку.
// * Зависимость мощности от задержки открытия триака расчитывается как интеграл от f до pi от sin(x)*sin(x)dx/(pi/2) и равен 1 - f / pi + sin(2 * f) / (2 * pi) (расчет нашел в нете). Расчет табличных значений приведен в таблице xls во вложении. 
// * Устанавливаем задержку на соответствующий таймер
int tableK[201]; //табличное значение отношения заданной мощности к текущей, умноженное на 1000
int tableT[201]; //табличное значение времени задежки открытия триака
unsigned int currentADC_AMPLITUDE_VOLTAGE; //мгновенное значение измеренного напряжения
  
void setup(){
  //Serial.begin(9600);//инициализация com1 для отладки
  // установка пинов
  pinMode(ZERO_CROSS, INPUT);     //определение направляения сигнала
  digitalWrite(ZERO_CROSS, HIGH); //включение подтягивающего резистора
  pinMode(TRIAC_GATE_PULSE, OUTPUT);      //определение направления сигнала
  
  //Расчет константы формулы определение соотношение заданной мощности к текущей, умноженной на 1000
  CONST = 2*heaterRESISTANCE/(Z*Z)*1000;

  tableK[200]=1000;  tableT[200]=1;
  tableK[199]=995;   tableT[199]=57;
  tableK[198]=990;   tableT[198]=72;
  tableK[197]=985;   tableT[197]=83;
  tableK[196]=980;   tableT[196]=92;
  tableK[195]=975;   tableT[195]=99;
  tableK[194]=970;   tableT[194]=106;
  tableK[193]=965;   tableT[193]=111;
  tableK[192]=960;   tableT[192]=117;
  tableK[191]=955;   tableT[191]=122;
  tableK[190]=950;   tableT[190]=126;
  tableK[189]=944;   tableT[189]=131;
  tableK[188]=940;   tableT[188]=135;
  tableK[187]=934;   tableT[187]=139;
  tableK[186]=930;   tableT[186]=142;
  tableK[185]=925;   tableT[185]=146;
  tableK[184]=920;   tableT[184]=149;
  tableK[183]=914;   tableT[183]=153;
  tableK[182]=910;   tableT[182]=156;
  tableK[181]=905;   tableT[181]=159;
  tableK[180]=900;   tableT[180]=162;
  tableK[179]=895;   tableT[179]=165;
  tableK[178]=889;   tableT[178]=168;
  tableK[177]=886;   tableT[177]=170;
  tableK[176]=880;   tableT[176]=173;
  tableK[175]=874;   tableT[175]=176;
  tableK[174]=871;   tableT[174]=178;
  tableK[173]=865;   tableT[173]=181;
  tableK[172]=861;   tableT[172]=183;
  tableK[171]=854;   tableT[171]=186;
  tableK[170]=850;   tableT[170]=188;
  tableK[169]=846;   tableT[169]=190;
  tableK[168]=840;   tableT[168]=193;
  tableK[167]=835;   tableT[167]=195;
  tableK[166]=831;   tableT[166]=197;
  tableK[165]=824;   tableT[165]=200;
  tableK[164]=819;   tableT[164]=202;
  tableK[163]=815;   tableT[163]=204;
  tableK[162]=810;   tableT[162]=206;
  tableK[161]=805;   tableT[161]=208;
  tableK[160]=800;   tableT[160]=210;
  tableK[159]=796;   tableT[159]=212;
  tableK[158]=791;   tableT[158]=214;
  tableK[157]=786;   tableT[157]=216;
  tableK[156]=781;   tableT[156]=218;
  tableK[155]=776;   tableT[155]=220;
  tableK[154]=770;   tableT[154]=222;
  tableK[153]=765;   tableT[153]=224;
  tableK[152]=760;   tableT[152]=226;
  tableK[151]=755;   tableT[151]=228;
  tableK[150]=749;   tableT[150]=230;
  tableK[149]=744;   tableT[149]=232;
  tableK[148]=741;   tableT[148]=233;
  tableK[147]=736;   tableT[147]=235;
  tableK[146]=730;   tableT[146]=237;
  tableK[145]=725;   tableT[145]=239;
  tableK[144]=719;   tableT[144]=241;
  tableK[143]=716;   tableT[143]=242;
  tableK[142]=711;   tableT[142]=244;
  tableK[141]=705;   tableT[141]=246;
  tableK[140]=699;   tableT[140]=248;
  tableK[139]=696;   tableT[139]=249;
  tableK[138]=691;   tableT[138]=251;
  tableK[137]=685;   tableT[137]=253;
  tableK[136]=679;   tableT[136]=255;
  tableK[135]=676;   tableT[135]=256;
  tableK[134]=670;   tableT[134]=258;
  tableK[133]=664;   tableT[133]=260;
  tableK[132]=661;   tableT[132]=261;
  tableK[131]=655;   tableT[131]=263;
  tableK[130]=649;   tableT[130]=265;
  tableK[129]=646;   tableT[129]=266;
  tableK[128]=640;   tableT[128]=268;
  tableK[127]=634;   tableT[127]=270;
  tableK[126]=631;   tableT[126]=271;
  tableK[125]=625;   tableT[125]=273;
  tableK[124]=619;   tableT[124]=275;
  tableK[123]=615;   tableT[123]=276;
  tableK[122]=609;   tableT[122]=278;
  tableK[121]=606;   tableT[121]=279;
  tableK[120]=600;   tableT[120]=281;
  tableK[119]=594;   tableT[119]=283;
  tableK[118]=591;   tableT[118]=284;
  tableK[117]=584;   tableT[117]=286;
  tableK[116]=581;   tableT[116]=287;
  tableK[115]=575;   tableT[115]=289;
  tableK[114]=569;   tableT[114]=291;
  tableK[113]=565;   tableT[113]=292;
  tableK[112]=559;   tableT[112]=294;
  tableK[111]=556;   tableT[111]=295;
  tableK[110]=549;   tableT[110]=297;
  tableK[109]=546;   tableT[109]=298;
  tableK[108]=540;   tableT[108]=300;
  tableK[107]=534;   tableT[107]=302;
  tableK[106]=530;   tableT[106]=303;
  tableK[105]=524;   tableT[105]=305;
  tableK[104]=521;   tableT[104]=306;
  tableK[103]=514;   tableT[103]=308;
  tableK[102]=511;   tableT[102]=309;
  tableK[101]=505;   tableT[101]=311;
  tableK[100]=502;   tableT[100]=312;
  tableK[99]=495;    tableT[99]=314;
  tableK[98]=489;    tableT[98]=316;
  tableK[97]=486;    tableT[97]=317;
  tableK[96]=479;    tableT[96]=319;
  tableK[95]=476;    tableT[95]=320;
  tableK[94]=470;    tableT[94]=322;
  tableK[93]=466;    tableT[93]=323;
  tableK[92]=460;    tableT[92]=325;
  tableK[91]=454;    tableT[91]=327;
  tableK[90]=451;    tableT[90]=328;
  tableK[89]=444;    tableT[89]=330;
  tableK[88]=441;    tableT[88]=331;
  tableK[87]=435;    tableT[87]=333;
  tableK[86]=431;    tableT[86]=334;
  tableK[85]=425;    tableT[85]=336;
  tableK[84]=419;    tableT[84]=338;
  tableK[83]=416;    tableT[83]=339;
  tableK[82]=409;    tableT[82]=341;
  tableK[81]=406;    tableT[81]=342;
  tableK[80]=400;    tableT[80]=344;
  tableK[79]=394;    tableT[79]=346;
  tableK[78]=391;    tableT[78]=347;
  tableK[77]=385;    tableT[77]=349;
  tableK[76]=381;    tableT[76]=350;
  tableK[75]=375;    tableT[75]=352;
  tableK[74]=369;    tableT[74]=354;
  tableK[73]=366;    tableT[73]=355;
  tableK[72]=360;    tableT[72]=357;
  tableK[71]=354;    tableT[71]=359;
  tableK[70]=351;    tableT[70]=360;
  tableK[69]=345;    tableT[69]=362;
  tableK[68]=339;    tableT[68]=364;
  tableK[67]=336;    tableT[67]=365;
  tableK[66]=330;    tableT[66]=367;
  tableK[65]=324;    tableT[65]=369;
  tableK[64]=321;    tableT[64]=370;
  tableK[63]=315;    tableT[63]=372;
  tableK[62]=309;    tableT[62]=374;
  tableK[61]=306;    tableT[61]=375;
  tableK[60]=301;    tableT[60]=377;
  tableK[59]=295;    tableT[59]=379;
  tableK[58]=289;    tableT[58]=381;
  tableK[57]=286;    tableT[57]=382;
  tableK[56]=281;    tableT[56]=384;
  tableK[55]=275;    tableT[55]=386;
  tableK[54]=270;    tableT[54]=388;
  tableK[53]=264;    tableT[53]=390;
  tableK[52]=259;    tableT[52]=392;
  tableK[51]=256;    tableT[51]=393;
  tableK[50]=251;    tableT[50]=395;
  tableK[49]=245;    tableT[49]=397;
  tableK[48]=240;    tableT[48]=399;
  tableK[47]=235;    tableT[47]=401;
  tableK[46]=230;    tableT[46]=403;
  tableK[45]=224;    tableT[45]=405;
  tableK[44]=219;    tableT[44]=407;
  tableK[43]=214;    tableT[43]=409;
  tableK[42]=209;    tableT[42]=411;
  tableK[41]=204;    tableT[41]=413;
  tableK[40]=200;    tableT[40]=415;
  tableK[39]=195;    tableT[39]=417;
  tableK[38]=190;    tableT[38]=419;
  tableK[37]=185;    tableT[37]=421;
  tableK[36]=181;    tableT[36]=423;
  tableK[35]=176;    tableT[35]=425;
  tableK[34]=169;    tableT[34]=428;
  tableK[33]=165;    tableT[33]=430;
  tableK[32]=160;    tableT[32]=432;
  tableK[31]=156;    tableT[31]=434;
  tableK[30]=150;    tableT[30]=437;
  tableK[29]=146;    tableT[29]=439;
  tableK[28]=139;    tableT[28]=442;
  tableK[27]=135;    tableT[27]=444;
  tableK[26]=129;    tableT[26]=447;
  tableK[25]=126;    tableT[25]=449;
  tableK[24]=120;    tableT[24]=452;
  tableK[23]=114;    tableT[23]=455;
  tableK[22]=111;    tableT[22]=457;
  tableK[21]=105;    tableT[21]=460;
  tableK[20]=100;    tableT[20]=463;
  tableK[19]=95;     tableT[19]=466;
  tableK[18]=90;     tableT[18]=469;
  tableK[17]=86;     tableT[17]=472;
  tableK[16]=80;     tableT[16]=476;
  tableK[15]=75;     tableT[15]=479;
  tableK[14]=70;     tableT[14]=483;
  tableK[13]=66;     tableT[13]=486;
  tableK[12]=60;     tableT[12]=490;
  tableK[11]=56;     tableT[11]=494;
  tableK[10]=50;     tableT[10]=499;
  tableK[9]=45;      tableT[9]=503;
  tableK[8]=40;      tableT[8]=508;
  tableK[7]=35;      tableT[7]=514;
  tableK[6]=30;      tableT[6]=519;
  tableK[5]=25;      tableT[5]=526;
  tableK[4]=20;      tableT[4]=533;
  tableK[3]=15;      tableT[3]=542;
  tableK[2]=10;      tableT[2]=552;
  tableK[1]=5;       tableT[1]=568;
  tableK[0]=1;       tableT[0]=610;

  ////////////////////////////
  //// установка Таймера3 ////
  ////////////////////////////  
  // Таймер3 используется для организации фазового управления напряжением (мощностью)
  // Последоватлеьность работы
  // * таймер установлен но запрещен
  // * обнаружен сигнал пересечения нуля на 2
  // * таймер начинает считать с нуля
  // * компаратор установлен на значение задержки включения триака
  // * счетчик таймера достигает значения компаратора
  // * обработчик прерывания компаратора включает триак (подачет сигнал на вывод GATE)
  // * счетчик таймера устанавливает длиительность импульса включения триака (разрядность счетчика минус длительность импульсов)
  // * счетчик таймера достигает переполнения
  // * обработчик прерывания по перевыполнению снимает управляющий сигнал с триака
  // * триак закрывается сам при следующем переходе через ноль (особенность работы триаков)
  // Частота процессора 16MHz. 
  // Использование делителя 256 обеспечивает длительность тика 16 мкс 16МГц/256=62500Гц = 16мкс/тик)
  // Частота полуволны переменнного напряжения 50Гц составляет 100Гц 
  // В течение полуволны таймер выдает 625 тиков (10мс)
  TIMSK3 &= (0<<OCIE3B); //1-разрешение прерывания по компаратору В
  TIMSK3 |= (1<<OCIE3A); //1-разрешение прерывания по компаратору А
  TIMSK3 |= (1<<TOIE3); //1-разрешение прерывания по переполнению    
  TCCR3A = 0x00;    //установка регистров контроля таймера (timer control register) для
  TCCR3B = 0x00;    //нормальной работы, таймер запрещен
  OCR3A = 550;     //инициализация значения компаратора

  ////////////////////////////
  //// установка Таймера1 ////
  ////////////////////////////  
  // Таймер1 задает частоту опроса АЦП
  // Частота процессора 16MHz. 
  // Использование делителя 256 обеспечивает длительность тика 16 мкс 16МГц/256=62500Гц = 16мкс/тик)
  // Частота полуволны переменнного напряжения 50Гц составляет 100Гц 
  // В течение полуволны таймер выдает 62500Гц/100Гц=625 тиков (10мс)
  // Для обеспечения 60 выборок в течение полуволны значние компаратора должно быть 625/60=10 тик (160 мксек)
  TIMSK1 &= (0<<OCIE1B); //1-разрешение прерывания по компаратору В
  TIMSK1 &= (0<<OCIE1A); //1-разрешение прерывания по компаратору А
  TIMSK1 |= (1<<TOIE1); //1-разрешение прерывания по переполнению  
  TCCR1A = 0x00;    //установка регистров контроля таймера (timer control register) для
  TCCR1B = 0x00;    //нормальной работы, таймер запрещен
  
  /////////////////////////// 
  //// Инициализация АЦП ////
  ///////////////////////////
//  DIDR0=0xFF; //запретить дискретные входы на аналоговых пинах 7-0
//  DIDR2=0xFF; //запретить дискретные входы на аналоговых пинах 15-8
 
  //инициализация мультиплексора
  //бит 7-6 выбор опорного напряжения 2,56В
  //бит 5   установка право ориентированного порядка следования старшего и младшего байт АЦП
  //бит 4-0 выбор аналогового пина 3
  ADMUX = 0b11000011;
  
  //инициализация регистра ADCSRA – ADC Control and Status Register A
  //бит 7 разшершить АЦП   
  //бит 6 запустить преобразование
  //бит 5 разрешить преобразование по сигналу триггера
  //бит 4 флаг прерывания
  //бит 3 разшерить прерывания АЦП
  //бит 2-0 выбор делителя частоты (64) 
  ADCSRA = 0b11101110;
  
  //Установка регистра ADCSRB – ADC Control and Status Register B 
  // бит 6 Analog Multiplecsor Comparator Enable
  // бит 3 выбор аналогового пина 3
  // бит 2-0 выбор триггера для АЦП преобразования - переполнение Таймера 1
  ADCSRB = 0b01000110;
  
  // инициализация прерывания при переходе нуля напряжения
  // прерывание по переднему фронту сигнала, привязанного к пину 2
  attachInterrupt(0,zeroCrossingInterrupt, RISING);    

  //разрешение прерываний
  sei();
}  

//Обработчик прерывания перехода нуля
void zeroCrossingInterrupt(){ //обнаружен переход нуля   
  TCCR3B=0x04; //запуск таймера 3 с делителем 256 (16МГц/256=62500Гц = 16мкс/тик)
  TCNT3 = 0;   //обнуление значения таймера 3
  
  if (statusVOLTAGE_MEASUREMENT == 1)  //если получен запрос на начало выборки измерения напряжения - начинаем выборку
  {
    TCCR1B=0x04; //запуск таймера 1 с делителем 256 (16МГц/256=62500Гц = 16мкс/тик)
    TCNT1 = 65536 - samples_frequencyVOLTAGE_MEASUREMENT; //устанавливаем значение таймера 1
    maxADC_AMPLITUDE_VOLTAGE = 0; //обнуляем максимальное амплитундое значение напряжения
    statusVOLTAGE_MEASUREMENT =2; //выборка значений напряжения полуволны в процессе
  } else
    {
      if (statusVOLTAGE_MEASUREMENT == 2)  //если идет процесс измерения напряжения и вызывается повторное прерывание по переходу нуля
      {                                     //останавливаем выборку и устанавливаем задержку открытия триака
        TCCR1B=0x00; //останавливаем таймер 1
        statusVOLTAGE_MEASUREMENT = 3;
        int i=0;
        K = (CONST * requiredPOWER/maxADC_AMPLITUDE_VOLTAGE/maxADC_AMPLITUDE_VOLTAGE); //расчитываем отношение заданной к текущей мощности
        if (K>=1000)i=200; //ищем индекс ближайщего значения К в таблице
        else
          if (K<5) i=0;
            else
            { 
              i= (K-6)/5; //для ускорения поиска по массиву               
              while (K >= tableK[i]) i++;
            }
        OCR3A = tableT[i]; //по найденному индексу устанавливаем заданное время задержки открытия триака
        statusVOLTAGE_MEASUREMENT = 1; //устанавливаем флаг старта новой выборки
        Serial.println(tableT[i]);
      }  
    }
}

ISR(TIMER3_COMPA_vect){ //обработчик прерывания по достижению таймером 3 значения компаратора А
  PORTA |= (1<<PORTA0);   //подача управляющего сигнала на вход GATE тиака (PIN22)  
  TCNT3 = 65536 - widthTRIAC_GATE_PULSE; //установка значения таймера на длительность импульса управляющего сигнала триака
                                       //управляющий сигнал будет отключен по прерыванию переполнения таймера
}

//Обработчик прерывания по переполнению таймера 3
ISR(TIMER3_OVF_vect){ 
  PORTA &= (0<<PORTA0);   //снятие управляющего сигнала с входа GATE тиака (PIN22)
  TCCR3B = 0x00;          //запрещение таймера для избежания ложных срабатываний
}

ISR(TIMER1_OVF_vect){ //обработчик прерывания по достижению таймером 3 значения компаратора B
  TCNT1 = 65536 - samples_frequencyVOLTAGE_MEASUREMENT;   //установка значения таймера 1 для следующего вызова прерывания по переполнению
}

//Обработчик прерывания окончания АЦП преобразования
//данный обработчик считывает значение текущего напряжения с аналогового пина 3 в массив выборок
ISR(ADC_vect){
  // измеренное значение амплитудного напряжения
  currentADC_AMPLITUDE_VOLTAGE = ADCL;//сохранение младшего байта АЦП
  currentADC_AMPLITUDE_VOLTAGE += ADCH << 8;//сохранение старшего байта АЦП
  if (maxADC_AMPLITUDE_VOLTAGE < currentADC_AMPLITUDE_VOLTAGE) maxADC_AMPLITUDE_VOLTAGE = currentADC_AMPLITUDE_VOLTAGE; //определение максимального амплитудного значения
}


void loop(){
  
    requiredPOWER = 30; // задаваемая мощность, в Ватах, для лампочки 75 Вт

}
